/*
**
* BEGIN_COPYRIGHT
*
* Copyright (C) 2008-2018 SciDB, Inc.
* All Rights Reserved.
*
* SciDB is free software: you can redistribute it and/or modify
* it under the terms of the AFFERO GNU General Public License as published by
* the Free Software Foundation.
*
* SciDB is distributed "AS-IS" AND WITHOUT ANY WARRANTY OF ANY KIND,
* INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,
* NON-INFRINGEMENT, OR FITNESS FOR A PARTICULAR PURPOSE. See
* the AFFERO GNU General Public License for the complete license terms.
*
* You should have received a copy of the AFFERO GNU General Public License
* along with SciDB.  If not, see <http://www.gnu.org/licenses/agpl-3.0.html>
*
* END_COPYRIGHT
*/


/**
 * @file DBArray.cpp
 *
 * @brief Database array implementation
 *
 * @author Konstantin Knizhnik <knizhnik@garret.ru>
 */
#include <log4cxx/logger.h>
#include <array/DBArray.h>
#include <smgr/io/InternalStorage.h>
#include <system/Exceptions.h>

namespace scidb
{
    static log4cxx::LoggerPtr logger(log4cxx::Logger::getLogger("scidb.services.storage"));
    //
    // DBArray
    //
    DBArray::DBArray(ArrayDesc const& desc, const std::shared_ptr<Query>& query)
    : _desc(desc)
    {
        LOG4CXX_DEBUG(logger, "DBArray::DBArray ID="<<_desc.getId()
                      <<", UAID="<<_desc.getUAId()
                      <<", ps="<<_desc.getDistribution()->getPartitioningSchema()
                      << ", desc="<< desc);
        _query = query;
        SCIDB_ASSERT(query);
        SCIDB_ASSERT(_desc.getDistribution()->getPartitioningSchema() != psUninitialized);
        SCIDB_ASSERT(_desc.getDistribution()->getPartitioningSchema() != psUndefined); // important for degraded mode, used by smgr
    }
    std::string const& DBArray::getName() const
    {
        return _desc.getName();
    }

    ArrayID DBArray::getHandle() const
    {
        return _desc.getId();
    }

    ArrayDesc const& DBArray::getArrayDesc() const
    {
        return _desc;
    }

    std::shared_ptr<ArrayIterator> DBArray::getIterator(AttributeID attId)
    {
       LOG4CXX_TRACE(logger, "Getting DB iterator for ID="<<_desc.getId()<<", attrID="<<attId);
       std::shared_ptr<Query> query(Query::getValidQueryPtr(_query));
       std::shared_ptr<const Array> arr= std::dynamic_pointer_cast<const Array>(shared_from_this());
       return StorageMan::getInstance().getArrayIterator(arr, attId, query);
    }

    std::shared_ptr<ConstArrayIterator> DBArray::getConstIterator(AttributeID attId) const
    {
        LOG4CXX_TRACE(logger, "Getting const DB iterator for ID="<<_desc.getId()<<", attrID="<<attId);
        std::shared_ptr<Query> query(Query::getValidQueryPtr(_query));
        std::shared_ptr<const Array> arr= std::dynamic_pointer_cast<const Array>(shared_from_this());
        return StorageMan::getInstance().getConstArrayIterator(arr, attId, query);
    }

    std::shared_ptr<CoordinateSet> DBArray::getChunkPositions() const
    {
        std::shared_ptr<Query> query(Query::getValidQueryPtr(_query));
        std::shared_ptr<CoordinateSet> result (new CoordinateSet());
        assert(query);
        StorageMan::getInstance().getChunkPositions(_desc, query, *(result.get()));
        return result;
    }

    void DBArray::removeDeadChunks(std::shared_ptr<Query>& query,
                                   std::set<Coordinates, CoordinatesLess> const& liveChunks)
    {
        StorageMan::getInstance().removeDeadChunks(_desc,
                                                   liveChunks,
                                                   query);
    }

    void DBArray::flush()
    {
        StorageMan::getInstance().flush();
    }


    void DBArray::removeVersions(std::shared_ptr<Query>& query, ArrayID lastLiveArrId)
    {
        StorageMan::getInstance().removeVersions(query->getQueryID(),
                                                 _desc.getUAId(),
                                                 lastLiveArrId);
    }

} // namespace scidb
