/*
**
* BEGIN_COPYRIGHT
*
* Copyright (C) 2008-2018 SciDB, Inc.
* All Rights Reserved.
*
* SciDB is free software: you can redistribute it and/or modify
* it under the terms of the AFFERO GNU General Public License as published by
* the Free Software Foundation.
*
* SciDB is distributed "AS-IS" AND WITHOUT ANY WARRANTY OF ANY KIND,
* INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,
* NON-INFRINGEMENT, OR FITNESS FOR A PARTICULAR PURPOSE. See
* the AFFERO GNU General Public License for the complete license terms.
*
* You should have received a copy of the AFFERO GNU General Public License
* along with SciDB.  If not, see <http://www.gnu.org/licenses/agpl-3.0.html>
*
* END_COPYRIGHT
*/

// de-facto std
#include <boost/numeric/conversion/cast.hpp>

// scidb
#include <query/Operator.h>
#include <query/OperatorLibrary.h>
#include <system/Exceptions.h>
#include <system/SystemCatalog.h>
#include <system/BlockCyclic.h>

#include <DLAErrors.h>

// local
#include "../scalapackFromCpp.hpp"

namespace scidb
{

// REFACTOR
inline bool hasSingleAttribute(ArrayDesc const& desc)
{
    return desc.getAttributes().size() == 1 ||
           (desc.getAttributes().size() == 2 && desc.getAttributes()[1].isEmptyIndicator());
}

// REFACTOR
///
/// integer rounds down to the nearest integer
/// this version of division rounds up to the nearest integer
/// like all division operators, it causes an exception if the divisor is zero
///
template<typename int_tt>
inline int_tt divCeil(int_tt val, int_tt divisor) {
    return (val + divisor - 1) / divisor ;
}

class MPIRankLogical: public LogicalOperator
{
public:
    MPIRankLogical(const std::string& logicalName, const std::string& alias):
        LogicalOperator(logicalName, alias)
    {
        ADD_PARAM_INPUT()
    }

    ArrayDesc inferSchema(std::vector<ArrayDesc> schemas, std::shared_ptr<Query> query);
};

ArrayDesc MPIRankLogical::inferSchema(std::vector<ArrayDesc> schemas, std::shared_ptr<Query> query)
{
    assert(schemas.size() == 1);

    if (!hasSingleAttribute(schemas[0]))
        throw PLUGIN_USER_EXCEPTION(DLANameSpace, SCIDB_SE_INFER_SCHEMA, DLA_ERROR2);
    if (schemas[0].getDimensions().size() != 2)
        throw PLUGIN_USER_EXCEPTION(DLANameSpace, SCIDB_SE_INFER_SCHEMA, DLA_ERROR3);
    if (schemas[0].getAttributes()[0].getType() != TID_DOUBLE)
        throw PLUGIN_USER_EXCEPTION(DLANameSpace, SCIDB_SE_INFER_SCHEMA, DLA_ERROR5);

    const Dimensions dims = schemas[0].getDimensions();
    DimensionDesc const& d0 = dims[0]; // kill these two off, to match most code?
    DimensionDesc const& d1 = dims[1];

    // must have finite size
    if (d0.isMaxStar() || d1.isMaxStar()) {
        throw PLUGIN_USER_EXCEPTION(DLANameSpace, SCIDB_SE_INFER_SCHEMA, DLA_ERROR9);
    }

    // is there a ctor that will take just the original names and do the following by a rule?
    Attributes atts(1);
    atts[0] = AttributeDesc(0, "left", TID_DOUBLE, 0, CompressorType::NONE);
    Dimensions outDims(2);
    outDims[0] = DimensionDesc(d0.getBaseName() + "_1",
                            d0.getStartMin(),
                            d0.getCurrStart(),
                            d0.getCurrEnd(),
                            d0.getEndMax(),
                            d0.getRawChunkInterval(),
                            0);
    outDims[1] = DimensionDesc(d0.getBaseName() + "_2",
                            d1.getStartMin(),
                            d1.getCurrStart(),
                            d1.getCurrEnd(),
                            d1.getEndMax(),
                            d1.getRawChunkInterval(),
                            0);

    ArrayDistPtr undefDist = ArrayDistributionFactory::getInstance()->construct(psUndefined,
                                                                                DEFAULT_REDUNDANCY);
    return ArrayDesc("mpirank", atts, outDims,
                     undefDist,
                     query->getDefaultArrayResidency());
}

REGISTER_LOGICAL_OPERATOR_FACTORY(MPIRankLogical, "_mpirank");

} // end namespace

